/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.bos.web; // NOPMD

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.openqa.selenium.By;
import org.openqa.selenium.NoSuchSessionException;
import org.openqa.selenium.SessionNotCreatedException;
import org.openqa.selenium.TimeoutException;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.firefox.FirefoxDriver;
import org.openqa.selenium.firefox.FirefoxOptions;
import org.openqa.selenium.firefox.FirefoxProfile;
import org.openqa.selenium.support.ui.ExpectedConditions;
import org.openqa.selenium.support.ui.WebDriverWait;

import com.gargoylesoftware.htmlunit.WebClient;

import de.frankmuenster.mahoe.Settings;
import de.frankmuenster.mahoe.statements.AbstractAccountStatementFinder;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.Utils;
import de.frankmuenster.mahoe.web.AbstractWebSite;
import de.frankmuenster.mahoe.web.MahoeKontoauszugPDF;
import de.frankmuenster.mahoe.web.SeleniumWebSite;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.rmi.AuslandsUeberweisung;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.jameica.messaging.StatusBarMessage;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * @author Frank Mnster
 */
public final class BosWebSite extends AbstractWebSite<BosWebAccountInformation, BosBankMessages>
        implements SeleniumWebSite {

    /**
     * I18N constant for BosWebSite.selDriverNotSet
     */
    private static final String I18N_BOS_WEB_SITE_SEL_DRIVER_NOT_SET = "BosWebSite.selDriverNotSet"; // NOPMD //$NON-NLS-1$

    /**
     * I18N constant BosWebSite.accountNumberNotSet
     */
    private static final String I18N_BOS_WEB_SITE_ACCOUNT_NUMBER_NOT_SET = "BosWebSite.accountNumberNotSet"; // NOPMD //$NON-NLS-1$

    /**
     * Mime type for downloads
     */
    private static final String MIME_TYPE_PDF = "application/pdf"; //$NON-NLS-1$

    /**
     * Our instance
     */
    private static BosWebSite webSite;

    /**
     * BOS runs with Selenium. This is the driver instance.
     */
    private WebDriver driver;

    /**
     * The account name
     */
    private final String userName;

    /**
     * Temporary folder to download the account sheets for further processing
     */
    private Path tmpDownloadFld;

    /**
     * Returns the instance of the BosWebsite. Creates a new one, if none exists
     * or the users nam eis different.
     *
     * @param konto
     *            the account to create the instance for
     * @throws WebSiteException
     * @throws ApplicationException
     * @return the instance of the BosWebsite for this account/user name
     */
    public static BosWebSite getInstance(final Konto konto) throws WebSiteException, ApplicationException {

        String userName = null;
        try {
            userName = konto.getKundennummer();
        } catch (final RemoteException e) {
            final String msg = I18N.tr("AbstractWebSite.loginMissing"); // NOPMD //$NON-NLS-1$
            Logger.error(msg, e);
            throw new ApplicationException(msg, e);
        }
        if (BosWebSite.webSite == null) {
            BosWebSite.webSite = new BosWebSite(konto, userName);
        } else {
            // Ist der Benutzername der selbe
            if (BosWebSite.webSite.userName.equals(userName)) {
                // Aktualisiere Startdatum der Umsatzabfrage und lsche
                // Umsatzliste
                final Date lastTransacDate = Utils.getLastUmsatzDate(konto, BosWebSite.webSite.getStartdate());
                BosWebSite.webSite.accountInfo.setLastTransactionDate(lastTransacDate);
            } else {
                BosWebSite.webSite = new BosWebSite(konto, userName);
            }
        }
        return BosWebSite.webSite;

    }

    /**
     * Creates a new WebSite instance to communicate with the bank
     *
     * @param konto
     * @param accountName
     * @throws WebSiteException
     */
    private BosWebSite(final Konto konto, final String accountName) throws WebSiteException {

        super(konto);
        this.userName = accountName;

    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#createSepaTransfer(de.willuhn.jameica.hbci.rmi.AuslandsUeberweisung)
     */
    @Override
    public void createSepaTransfer(final AuslandsUeberweisung aUeberweisung)
            throws WebSiteException, ApplicationException {

        // Will not be supported
    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#getAccountInformation()
     */
    @Override
    public BosWebAccountInformation getAccountInformation() throws WebSiteException, ApplicationException {

        if (this.driver == null) {
            // Inform user
            driverNotSendMsg();
            throw new ApplicationException(I18N.tr(BosWebSite.I18N_BOS_WEB_SITE_SEL_DRIVER_NOT_SET));
        } else {
            try {
                final String accountNumber = this.accountInfo.getKonto().getKontonummer();
                return BosWebSiteHelper.getAccountInfo(this.driver, this.accountInfo, accountNumber);
            } catch (final RemoteException e) {
                Logger.error(I18N.tr(BosWebSite.I18N_BOS_WEB_SITE_ACCOUNT_NUMBER_NOT_SET));
                throw new ApplicationException(I18N.tr(BosWebSite.I18N_BOS_WEB_SITE_ACCOUNT_NUMBER_NOT_SET), e);
            }
        }
    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#getAccountSheets(de.frankmuenster.mahoe.web.WebAccountInformation)
     */
    @Override
    public List<MahoeKontoauszugPDF> getAccountSheets(final BosWebAccountInformation accountInfo)
            throws WebSiteException {

        if (this.driver == null) {
            // Inform user
            driverNotSendMsg();
            throw new WebSiteException(I18N.tr(BosWebSite.I18N_BOS_WEB_SITE_SEL_DRIVER_NOT_SET),
                    WebSiteExceptionTypes.SERVICE_UNAVAILABLE);
        } else {

            final AbstractAccountStatementFinder<
                    BosWebAccountInformation> accountSheets = new BosAccountStatementFinder(accountInfo, this);
            accountSheets.findStatements();

            return accountSheets.getSheets();
        }
    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#getBookedTransactions()
     */
    @Override
    public List<Umsatz> getBookedTransactions() throws WebSiteException, ApplicationException {

        if (this.driver == null) {
            // Inform user
            driverNotSendMsg();
            throw new WebSiteException(I18N.tr(BosWebSite.I18N_BOS_WEB_SITE_SEL_DRIVER_NOT_SET),
                    WebSiteExceptionTypes.SERVICE_UNAVAILABLE);
        } else {
            try {
                final String accountNumber = this.accountInfo.getKonto().getKontonummer();
                return BosWebSiteHelper.getBookedTransactions(this.driver, this.accountInfo, accountNumber);
            } catch (final RemoteException e) {
                Logger.error(I18N.tr(BosWebSite.I18N_BOS_WEB_SITE_ACCOUNT_NUMBER_NOT_SET));
                throw new ApplicationException(I18N.tr(BosWebSite.I18N_BOS_WEB_SITE_ACCOUNT_NUMBER_NOT_SET), e);
            }
        }
    }

    /**
     * Returns the temporary folder to use for downloading account sheets.
     * Create a new one per session.
     *
     * @return the tmp folder to use during this session
     * @throws ApplicationException
     *             if the tmp folder cannot be created
     */
    @Override
    public Path getDownloadsDirectory() throws ApplicationException {

        synchronized (this) {

            if (this.tmpDownloadFld == null) {

                try {
                    this.tmpDownloadFld = Files.createTempDirectory("selleniumdownloads_"); //$NON-NLS-1$
                } catch (final IOException ex) {
                    throw new ApplicationException(I18N.tr("BosWebSite.failedToCreateTmpFolder"), ex); //$NON-NLS-1$
                }
            }
        }
        return this.tmpDownloadFld;
    }

    /**
     * Returns the actual driver
     *
     * @return the driver
     */
    @Override
    public WebDriver getDriver() {

        return this.driver;
    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#getPreBookedTransactions()
     */
    @Override
    public List<Umsatz> getPreBookedTransactions() throws WebSiteException, ApplicationException {

        // Doesn't exists by BOS
        return new ArrayList<>();
    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#isStillLoggedIn()
     */
    @Override
    public boolean isStillLoggedIn() throws WebSiteException {

        if (this.driver == null) {
            // Inform user
            driverNotSendMsg();
            throw new WebSiteException(I18N.tr(BosWebSite.I18N_BOS_WEB_SITE_SEL_DRIVER_NOT_SET),
                    WebSiteExceptionTypes.SERVICE_UNAVAILABLE);
        } else {
            String check = StringUtils.EMPTY;
            final WebDriverWait wait = new WebDriverWait(this.driver, 1);
            try {
                this.driver.switchTo().defaultContent();
                wait.until(ExpectedConditions.frameToBeAvailableAndSwitchToIt(By.name("frame_top"))); //$NON-NLS-1$
                wait.until(ExpectedConditions.elementToBeClickable(By.name("btnContinue"))); //$NON-NLS-1$
                check = this.driver.getPageSource();
            } catch (TimeoutException | NoSuchSessionException | SessionNotCreatedException e) {
                Logger.debug("Timeout aufgetretten --> also nicht angemeldet."); //$NON-NLS-1$
            }
            // Check if we are logged in
            return check.contains("Abmelden"); //$NON-NLS-1$
        }
    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#login(char[], java.lang.String[])
     */
    @Override
    public void login(final char[] password, final String... loginParams) throws WebSiteException {

        if (this.driver == null) {
            // Inform user
            driverNotSendMsg();
            throw new WebSiteException(I18N.tr(BosWebSite.I18N_BOS_WEB_SITE_SEL_DRIVER_NOT_SET),
                    WebSiteExceptionTypes.SERVICE_UNAVAILABLE);
        } else {

            Logger.debug("Starte mit Login fr Benutzername {0}" + this.userName); //$NON-NLS-1$

            if (loginParams.length != 0) {
                throw new WebSiteException(I18N.tr("AbstractBankWebSite.loginParameterNumberWrong"), //$NON-NLS-1$
                        WebSiteExceptionTypes.LOGIN_ERROR);
            }
            // check if driver is still active

            try {
                this.driver.getTitle();
            } catch (final NoSuchSessionException | SessionNotCreatedException e) {
                this.getWebClient();
            }
            // Perform login
            BosWebSiteHelper.login(this.driver, this.userName, java.nio.CharBuffer.wrap(password));
        }

    }

    /**
     * @see de.frankmuenster.mahoe.web.WebSite#logoff()
     */
    @Override
    public void logoff() throws WebSiteException {

        if (this.driver == null) {
            // Inform user
            driverNotSendMsg();
            throw new WebSiteException(I18N.tr(BosWebSite.I18N_BOS_WEB_SITE_SEL_DRIVER_NOT_SET),
                    WebSiteExceptionTypes.SERVICE_UNAVAILABLE);
        } else {
            BosWebSiteHelper.logoff(this.driver);

            // End driver
            this.driver.quit();

            // delete tmp folder
            try {
                final Path dir = this.getDownloadsDirectory();
                FileUtils.deleteDirectory(dir.toFile());
            } catch (final IOException | ApplicationException e) {
                Logger.error(I18N.tr("BosWebSite.faileToDeleteTmpFolder"), e); //$NON-NLS-1$
            }
        }
    }

    /**
     * @see de.frankmuenster.mahoe.web.AbstractWebSite#getStartdate()
     */
    @Override
    protected Calendar getStartdate() {

        // Setzte das startDatum auf Januar 2016
        final Calendar startCalendar = Calendar.getInstance();
        startCalendar.set(2016, 0, 1);
        return startCalendar;
    }

    /**
     * But we are using it also to get the Selenium driver.
     *
     * @see de.frankmuenster.mahoe.web.AbstractWebSite#getWebClient()
     */
    @Override
    protected WebClient getWebClient() {

        String tmpdir = StringUtils.EMPTY;
        try {
            tmpdir = this.getDownloadsDirectory().toString();
        } catch (final ApplicationException e) {
            Logger.error(I18N.tr("BosWebSite.failedToCreateTmpFolder"), e); //$NON-NLS-1$
        }

        // New Firefox option
        final FirefoxOptions options = new FirefoxOptions();

        // new profile with auto download PDF
        final FirefoxProfile fxProfile = new FirefoxProfile();
        fxProfile.setPreference("browser.download.folderList", 2); //$NON-NLS-1$
        fxProfile.setPreference("browser.download.manager.showWhenStarting", false); //$NON-NLS-1$
        fxProfile.setPreference("browser.download.dir", tmpdir); //$NON-NLS-1$
        fxProfile.setPreference("browser.download.useDownloadDir", true); //$NON-NLS-1$
        fxProfile.setPreference("browser.helperApps.neverAsk.saveToDisk", BosWebSite.MIME_TYPE_PDF); //$NON-NLS-1$
        fxProfile.setPreference("browser.helperApps.alwaysAsk.force", false); //$NON-NLS-1$
        fxProfile.setPreference("pdfjs.disabled", true); //$NON-NLS-1$
        fxProfile.setPreference("pdfjs.enabledCache.state", false); //$NON-NLS-1$

        options.setProfile(fxProfile);

        // Set the path to the Selenium driver
        final String selDriver = Settings.getSeleniumDriver();
        final Path driverPath = Paths.get(selDriver);
        if (StringUtils.isNotEmpty(selDriver) && Files.exists(driverPath)) {
            // Instantiate a new Firefox driver
            System.setProperty("webdriver.gecko.driver", selDriver); //$NON-NLS-1$
            this.driver = new FirefoxDriver(options);
        } else {
            // Inform user
            driverNotSendMsg();
        }

        // We don't have a WebClient, so it can be Null
        return null;
    }

    /**
     * Send status bar message that driver is not configured
     */
    private void driverNotSendMsg() {
        final StatusBarMessage msg = new StatusBarMessage(I18N.tr("BosWebSite.selDriverNotSetMsg"), //$NON-NLS-1$
                StatusBarMessage.TYPE_ERROR);
        Application.getMessagingFactory().sendMessage(msg);
    }
}
