/*******************************************************************************
 * Copyright 2015 - 2019 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.bos.web;

import java.rmi.RemoteException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.time.LocalDate;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;
import org.openqa.selenium.By;
import org.openqa.selenium.Keys;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.support.ui.ExpectedConditions;
import org.openqa.selenium.support.ui.Select;
import org.openqa.selenium.support.ui.WebDriverWait;

import de.frankmuenster.mahoe.banks.bos.BankOfScotland;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.Utils;
import de.frankmuenster.mahoe.utils.WebUtils;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.Settings;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.jameica.messaging.StatusBarMessage;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;

/**
 * @author Frank Mnster
 */
public final class BosWebSiteHelper {

    /**
     * Element with name btnContinue
     */
    private static final String ELEMENT_NAME_BTN_CONTINUE = "btnContinue"; //$NON-NLS-1$

    /**
     * HTML tag for the first cell
     */
    private static final String HTML_TAG_TD1 = WebUtils.TAG_TD + "[1]"; //$NON-NLS-1$

    /**
     * HTML link text to close something
     */
    private static final String LINK_TEXT_CLOSE = "close"; //$NON-NLS-1$

    /**
     * ID for the table with account and transaction information
     */
    private static final String GRAPHTABLE = "graphtable"; //$NON-NLS-1$

    /**
     * ID of the main frame
     */
    private static final String FRAME_TXN = "frame_txn"; //$NON-NLS-1$

    /**
     * ID of the top frame
     */
    private static final String FRAME_TOP = "frame_top"; //$NON-NLS-1$

    /**
     * Load additional account info, like IBAn, BIC and current saldo
     *
     * @param driver
     *            driver with logged in user
     * @param accountInfo
     *            the {@link BosWebAccountInformation} to add the data too
     * @param accountNumber
     *            the account number
     * @return an {@link BosWebAccountInformation} object with the loaded information
     */
    public static BosWebAccountInformation getAccountInfo(final WebDriver driver,
            final BosWebAccountInformation accountInfo, final String accountNumber) {

        // Open the account list
        BosWebSiteHelper.openAccountList(driver);

        // Search table
        final WebElement table = driver.findElement(By.id(BosWebSiteHelper.GRAPHTABLE));

        // and now get and loop over rows
        final List<WebElement> rows = table.findElements(By.tagName(WebUtils.TAG_TR));
        for (final WebElement row : rows) {
            final String cellText = row.findElement(By.xpath(BosWebSiteHelper.HTML_TAG_TD1)).getText();
            // Correct account?
            if (cellText.equalsIgnoreCase(accountNumber) && row.findElements(By.tagName(WebUtils.TAG_TD)).size() > 5) {
                // IBAN
                accountInfo.setIban(row.findElement(By.xpath("td[3]")).getText()); //$NON-NLS-1$
                // BIC
                accountInfo.setBic(row.findElement(By.xpath("td[4]")).getText()); //$NON-NLS-1$
                // balance
                final String balance = row.findElement(By.xpath("td[6]")).getText().substring(2); //$NON-NLS-1$
                accountInfo.setSaldo(Utils.convertStringToBigDecimal(balance));
                // stop looping over exiting accounts
                break;
            }
        }

        return accountInfo;
    }

    /**
     * Logins in the bank web page.
     *
     * @param driver
     *            to use to perform the login
     * @param accountName
     *            to use
     * @param password
     *            to use
     * @throws WebSiteException
     *             is thrown when something went wrong.
     */
    public static void login(final WebDriver driver, final String accountName, final CharSequence password)
            throws WebSiteException {

        driver.get(BankOfScotland.BOS_LOGIN);
        Logger.debug("Login page source:\n" + driver.getPageSource()); //$NON-NLS-1$

        // Switch to login frame
        driver.switchTo().frame(BosWebSiteHelper.FRAME_TXN);

        // Select and fill login fields
        final WebElement userId = driver.findElement(By.name("fldLoginUserId")); //$NON-NLS-1$
        userId.sendKeys(accountName); // $NON-NLS-1$

        final WebElement fldPassword = driver.findElement(By.name("fldPassword")); //$NON-NLS-1$
        fldPassword.sendKeys(password); // $NON-NLS-1$

        // Wait a bit as it seems the password is not written immediately
        Utils.waitIgnoreThreadInterrupts(300);

        final WebElement btnContinue = driver.findElement(By.name(BosWebSiteHelper.ELEMENT_NAME_BTN_CONTINUE));
        btnContinue.click();

        // Wait a bit as page is slow and then go ahead with security question
        Utils.waitIgnoreThreadInterrupts(500);
        final WebDriverWait wait = new WebDriverWait(driver, 10);
        wait.until(ExpectedConditions.elementToBeClickable(By.name("fldGO"))); //$NON-NLS-1$

        final WebElement answer1 = driver.findElement(By.name("fldanswer1")); //$NON-NLS-1$
        // get security question answer
        final CharSequence secPw = java.nio.CharBuffer.wrap(BosWebSiteHelper.getSecurityAnswer(driver));
        answer1.sendKeys(secPw); // $NON-NLS-1$
        // secPw = null; // NOPMD
        final WebElement fldGo = driver.findElement(By.name("fldGO")); //$NON-NLS-1$
        fldGo.click();

        // Again give page some time to start reloading
        Utils.waitIgnoreThreadInterrupts(500);
        driver.switchTo().defaultContent();
        wait.until(ExpectedConditions.frameToBeAvailableAndSwitchToIt(By.name(BosWebSiteHelper.FRAME_TOP)));

        wait.until(ExpectedConditions.elementToBeClickable(By.name(BosWebSiteHelper.ELEMENT_NAME_BTN_CONTINUE)));
        final String check = driver.getPageSource();

        // Check if we are logged in
        if (!check.contains("Abmelden")) { //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("BosWebSiteHelper.loginFailed"), WebSiteExceptionTypes.LOGIN_FAILED); //$NON-NLS-1$
        }

    }

    /**
     * Log off from bank page
     *
     * @param driver
     *            with logged in page
     * @throws WebSiteException
     */
    public static void logoff(final WebDriver driver) throws WebSiteException {

        driver.switchTo().defaultContent();
        driver.switchTo().frame(BosWebSiteHelper.FRAME_TOP);

        final WebElement logoff = driver.findElement(By.name(BosWebSiteHelper.ELEMENT_NAME_BTN_CONTINUE));
        logoff.click();
    }

    /**
     * Gets answer for the security question from the user
     *
     * @param driver
     *            with security question page open
     * @return the security question answer
     * @throws WebSiteException
     */
    private static char[] getSecurityAnswer(final WebDriver driver) throws WebSiteException {

        final WebElement fldquestion1 = driver.findElement(By.name("fldquestion1")); //$NON-NLS-1$
        final String question = fldquestion1.getAttribute("value"); //$NON-NLS-1$

        try {
            final String answer = Application.getCallback().askUser(question,
                    I18N.tr("BosWebSiteHelper.securityQuestion")); //$NON-NLS-1$
            return answer.toCharArray();
        } catch (final Exception e) { // NOPMD
            throw new WebSiteException(I18N.tr("BosWebSiteHelper.failedToOpenSecurityQuestionDialog"), //$NON-NLS-1$
                    WebSiteExceptionTypes.LOGIN_ERROR, e);
        }

    }

    private BosWebSiteHelper() {

        // this is a utility class
    }

    /**
     * @param driver
     * @param accountInfo
     * @param accountNumber
     * @return the list of transactions
     * @throws WebSiteException
     */
    public static List<Umsatz> getBookedTransactions(final WebDriver driver, final BosWebAccountInformation accountInfo,
            final String accountNumber) throws WebSiteException {

        final List<Umsatz> transactions = new ArrayList<>();

        // get from date
        final String fromDate = BosWebSiteHelper.caclulateFromDate(accountInfo);

        try {

            // And and open transaction from last start days
            BosWebSiteHelper.findAndOpenAccount(driver, accountNumber);

            // open transaction from last year
            BosWebSiteHelper.openFirstTransactiosnPage(driver, fromDate);

            // Check if table exists
            final WebDriverWait wait = new WebDriverWait(driver, 10);
            wait.until(ExpectedConditions.or(ExpectedConditions.presenceOfElementLocated(By.name("pageform")), //$NON-NLS-1$
                    ExpectedConditions.presenceOfElementLocated(By.id("ui-dialog-title-dialog-modal-error")))); //$NON-NLS-1$

            // Check for popup and close
            final List<WebElement> errorDialog = driver.findElements(By.linkText(BosWebSiteHelper.LINK_TEXT_CLOSE));
            if (errorDialog.isEmpty()) {
                // add all booked transactions
                transactions.addAll(BosWebSiteHelper.readPage(driver));
            } else {
                // Get error text
                final WebElement errorTxt = driver.findElement(By.id("dialog-modal-error")); //$NON-NLS-1$
                errorDialog.get(0).click();
                Logger.warn(errorTxt.getText());
                final StatusBarMessage msg = new StatusBarMessage(
                        I18N.tr("BosWebSiteHelper.failedToOpenTransactionPage"), StatusBarMessage.TYPE_ERROR); //$NON-NLS-1$
                Application.getMessagingFactory().sendMessage(msg);

            }
        } catch (final RemoteException e) {
            throw new WebSiteException(I18N.tr("BosWebSiteHelper.failedToLoadTransactions"), e); //$NON-NLS-1$
        }

        return transactions;
    }

    /**
     * Reads all booked transactions from the given page
     *
     * @param driver
     *            with the page with the table of transactions loaded
     * @return a list with all transactions
     * @throws RemoteException
     *             if the creation of the transaction in the databases fails
     */
    private static List<Umsatz> readPage(final WebDriver driver) throws RemoteException {

        final HBCIDBService dbService = Settings.getDBService();
        final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy", //$NON-NLS-1$
                Locale.getDefault());

        final List<Umsatz> transactions = new ArrayList<>();

        final List<WebElement> tables = driver.findElements(By.className(BosWebSiteHelper.GRAPHTABLE));
        if (tables.size() == 2) { // NOPMD
            final List<WebElement> rows = tables.get(1).findElements(By.tagName(WebUtils.TAG_TR));
            for (final WebElement row : rows) {
                final String cell1 = row.findElement(By.xpath(BosWebSiteHelper.HTML_TAG_TD1)).getText();
                if ("Buchung".equals(cell1)) { //$NON-NLS-1$
                    continue;
                }

                final Umsatz transaction = dbService.createObject(Umsatz.class, null);

                // Buchungsdatum
                Date date = null;
                try {
                    date = sdf.parse(cell1);
                } catch (final ParseException e) {
                    Logger.error(I18N.tr("BceeWebSiteTransactionsImpl.parseBookingDateFailed", //$NON-NLS-1$
                            e.getMessage()), e);
                }
                transaction.setDatum(date);

                // Valuta
                try {
                    date = sdf.parse(row.findElement(By.xpath("td[2]")).getText()); //$NON-NLS-1$
                } catch (final ParseException e) {
                    Logger.error(I18N.tr("BceeWebSiteTransactionsImpl.parseValutaDateFailed", //$NON-NLS-1$
                            e.getMessage()), e);
                }
                transaction.setValuta(date);

                // Name
                final String name = row.findElement(By.xpath("td[3]")).getText(); //$NON-NLS-1$
                transaction.setGegenkontoName(name);

                // Zweck
                final String zweck = row.findElement(By.xpath("td[4]")).getText(); //$NON-NLS-1$
                Utils.loadUmsatz(transaction, zweck);

                String betrag = row.findElement(By.xpath("td[6]")).getText(); //$NON-NLS-1$
                if (StringUtils.isEmpty(betrag)) {
                    betrag = "-" + row.findElement(By.xpath("td[5]")).getText();  //$NON-NLS-1$//$NON-NLS-2$
                }
                transaction.setBetrag(Utils.convertStringToDouble(betrag));

                transactions.add(transaction);

            }

        }

        return transactions;
    }

    /**
     * Open the account with the given account number.
     *
     * @param driver
     *            the driver with logged in BOS account
     * @param accountNumber
     *            the number of the account to open
     */
    public static void findAndOpenAccount(final WebDriver driver, final String accountNumber) {

        BosWebSiteHelper.openAccountList(driver);

        // Search table
        final WebElement table = driver.findElement(By.id(BosWebSiteHelper.GRAPHTABLE));

        // and now get and loop over rows
        final List<WebElement> rows = table.findElements(By.tagName(WebUtils.TAG_TR));
        for (final WebElement row : rows) {
            final String cellText = row.findElement(By.xpath(BosWebSiteHelper.HTML_TAG_TD1)).getText();
            // Correct account?
            if (cellText.equalsIgnoreCase(accountNumber) && row.findElements(By.tagName(WebUtils.TAG_TD)).size() > 5) {
                // open account
                final WebElement link = row.findElement(By.xpath("td[1]/div/a")); //$NON-NLS-1$
                link.click();
            }
        }
    }

    /**
     * Opens the account list of the logged in user
     *
     * @param driver
     */
    private static void openAccountList(final WebDriver driver) {

        // Switch to top frame and reload 'Kontoinformationen'
        driver.switchTo().defaultContent();
        driver.switchTo().frame(BosWebSiteHelper.FRAME_TOP);
        WebElement accInfoLink = driver.findElement(By.xpath("/html/body/form[1]/div/div/div[3]/ul/li[1]/a")); //$NON-NLS-1$
        accInfoLink.click();

        final WebDriverWait wait = new WebDriverWait(driver, 10);
        driver.switchTo().defaultContent();
        wait.until(ExpectedConditions.frameToBeAvailableAndSwitchToIt(By.name(BosWebSiteHelper.FRAME_TXN)));

        wait.until(
                ExpectedConditions.or(ExpectedConditions.presenceOfElementLocated(By.id(BosWebSiteHelper.GRAPHTABLE)),
                        ExpectedConditions.presenceOfElementLocated(By.id("ui-dialog-title-dialog-modal-error")))); //$NON-NLS-1$

        // Check error dialog and try reload
        final List<WebElement> errorDialog = driver.findElements(By.linkText(BosWebSiteHelper.LINK_TEXT_CLOSE));
        if (!errorDialog.isEmpty()) {
            errorDialog.get(0).click();
            // Switch to top frame and reload 'Kontoinformationen'
            driver.switchTo().defaultContent();
            driver.switchTo().frame(BosWebSiteHelper.FRAME_TOP);
            accInfoLink = driver.findElement(By.xpath("/html/body/form[1]/div/div/div[3]/ul/li[1]/a")); //$NON-NLS-1$
            accInfoLink.click();

            driver.switchTo().defaultContent();
            wait.until(ExpectedConditions.frameToBeAvailableAndSwitchToIt(By.name(BosWebSiteHelper.FRAME_TXN)));
            wait.until(ExpectedConditions.presenceOfElementLocated(By.id(BosWebSiteHelper.GRAPHTABLE)));

        }

    }

    private static void openFirstTransactiosnPage(final WebDriver driver, final String fromDate) {

        final WebElement transactionBtn = driver.findElement(
                By.xpath("/html/body/form/table/tbody/tr/td/div[2]/table/tbody/tr[2]/td/div/div/div/ul/li[2]/input")); //$NON-NLS-1$
        transactionBtn.click();

        final WebDriverWait wait = new WebDriverWait(driver, 10);
        wait.until(ExpectedConditions.presenceOfElementLocated(By.id("fldsearch1"))); //$NON-NLS-1$

        // Check for popup and close
        final List<WebElement> errorDialog = driver.findElements(By.linkText(BosWebSiteHelper.LINK_TEXT_CLOSE));
        if (!errorDialog.isEmpty()) {
            errorDialog.get(0).click();
        }

        final Select dateRage = new Select(driver.findElement(By.id("fldsearch1"))); //$NON-NLS-1$
        final WebElement fromDateFld = driver.findElement(By.name("fldfromdate")); //$NON-NLS-1$
        // Select "Gewhlter Zeitraum"
        dateRage.selectByValue("3"); //$NON-NLS-1$

        // clean from date field
        for (int i = 0; i < 12; i++) {
            fromDateFld.sendKeys(Keys.BACK_SPACE);
        }
        // and set from date
        fromDateFld.sendKeys(fromDate);
        Logger.debug("fromDate:  " + fromDateFld.getAttribute("value")); //$NON-NLS-1$//$NON-NLS-2$

        // Now load transactions
        final WebElement submit = driver.findElement(By.name("fldsubmit")); //$NON-NLS-1$
        submit.click();
    }

    /**
     * Calculate the max. days to retrieve transaction. Either the lastTransaction Date from accountInfo or the last
     * 365.
     *
     * @param accountInfo
     * @return the date to start loading transactions from as string.
     */
    private static String caclulateFromDate(final BosWebAccountInformation accountInfo) {

        LocalDate startdate;
        final LocalDate date365 = LocalDate.now().minusDays(364);
        if (accountInfo.getLastTransactionDate() == null) {
            startdate = date365;
        } else {
            final Date date = accountInfo.getLastTransactionDate();
            if (date instanceof java.sql.Date) {
                startdate = ((java.sql.Date) date).toLocalDate();
            } else {
                startdate = date.toInstant().atZone(ZoneId.systemDefault()).toLocalDate();
            }
            if (startdate.isBefore(date365)) {
                startdate = date365;
            }
        }

        final StringBuilder fromDate = new StringBuilder(10);
        fromDate.append(startdate.getDayOfMonth()).append('.').append(startdate.getMonthValue()).append('.')
                .append(startdate.getYear());

        return fromDate.toString();

    }

}
