/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they
 * will be approved by the European Commission - subsequent
 * versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the
 * Licence.
 * You may obtain a copy of the Licence at:
 *
 *    https://progs.frankmuenster.de/eupl_v1-2_en/
 *
 * Unless required by applicable law or agreed to in
 * writing, software distributed under the Licence is
 * distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied.
 * See the Licence for the specific language governing
 * permissions and limitations under the Licence.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.santander.clearing;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.rmi.RemoteException;
import java.time.LocalDate;
import java.util.Date;

import org.apache.commons.lang.NullArgumentException;
import org.apache.commons.lang.StringUtils;

import de.frankmuenster.mahoe.MahoeSynchronizeBackend;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.Utils;
import de.willuhn.datasource.rmi.DBIterator;
import de.willuhn.jameica.hbci.Settings;
import de.willuhn.jameica.hbci.rmi.AuslandsUeberweisung;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.jameica.messaging.StatusBarMessage;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Bean mit Ausgleichbuchungsinformationen
 *
 * @author Frank Mnster
 */
public class ClearingEntry {

    /**
     * Der Ausgleichsbetrag
     */
    private final BigDecimal betrag;

    /**
     * Das F&auml;lligkitsdatum
     */
    private final Date datum;

    /**
     * Die Kreditkartennummer
     */
    private final String kartenNummer;

    /**
     * Der Karteninhaber
     */
    private final String inhaber;

    /**
     * Die IBAN
     */
    private final String iban;

    /**
     * Die BIC
     */
    private final String bic;

    /**
     * Das Abbuchungskonto
     */
    private Konto konto;

    /**
     * Erstellt eine neue Ausgleichsbuchung mit den angegebenen Daten.
     *
     * @param kartenNumber
     * @param inhaber
     * @param iban
     * @param bic
     * @param datum
     * @param betrag
     * @throws NullArgumentException
     */
    public ClearingEntry(final String kartenNumber, final String inhaber, final String iban, final String bic,
	    final Date datum, final BigDecimal betrag) throws NullArgumentException {
	this.kartenNummer = kartenNumber;
	this.inhaber = inhaber;
	this.iban = iban;
	this.bic = bic;
	this.betrag = betrag.setScale(2, RoundingMode.HALF_UP);
	this.datum = datum;
	// Alle Felder gesetzt
	this.checkParameter();
    }

    /**
     * berprpt, ob eine Ausgleichsbuchung in den letzten 365 Tagen mit den
     * selben Betrag schon existiert. Es werden alle Hibiscus-Konten durchsucht.
     *
     * @return <i>true</i>, wenn die Ausgleichbuchung schon existiert,
     *         <i>false</i> sonst.
     */
    public boolean exists() {

	boolean exists = false;

	try {
	    final HBCIDBService hcbiService = Settings.getDBService();
	    // Erst mal die schon abgefragten Umstze
	    exists = this.checkTransactions(hcbiService);

	    if (!exists) {
		// Dann noch die SEPA-berweisungen
		exists = this.checkSEPA(hcbiService);
	    }
	} catch (final RemoteException e) {
	    // Auf true setzen, damit die berweisung nicht ggf. doppelt
	    // angelegt wird.
	    exists = true;
	    final String msg = I18N.tr("ClearingEntry.searchClearingTransactionFailed"); //$NON-NLS-1$
	    Logger.error(msg, e);
	    final StatusBarMessage msgUser = new StatusBarMessage(msg, StatusBarMessage.TYPE_ERROR);
	    Application.getMessagingFactory().sendMessage(msgUser);
	}
	return exists;
    }

    /**
     * Erstellt die &Uuml;berweisung in Hibiscus. Passiert schon mal nur, wenn
     * der angegeben Betrag kleiner 0 ist.
     *
     * @return <i>true</i> falls die &Uuml;berweisung angelegt werden konnte,
     *         <i>false</i> sonst und wenn keine Expection ausgel&ouml;st wurde.
     * @throws NullArgumentException
     */
    public boolean saveEntry() throws NullArgumentException {

	if (this.konto == null) {
	    throw new NullArgumentException(I18N.tr("ClearingEntry.clearingAccountIsNull")); //$NON-NLS-1$
	}
	boolean saved = false;
	if (this.betrag.signum() < 0) {
	    try {
		final AuslandsUeberweisung clearingTransac = (AuslandsUeberweisung) Settings.getDBService()
			.createObject(AuslandsUeberweisung.class, null);
		clearingTransac.setKonto(this.konto);
		// Umlaute ersetzten
		clearingTransac.setGegenkontoName(Utils.replaceUmlauts(this.inhaber));
		clearingTransac.setGegenkontoNummer(this.iban);
		clearingTransac.setGegenkontoBLZ(this.bic);
		final Double myBetrag = this.betrag.negate().doubleValue();
		if (!myBetrag.isNaN()) {
		    clearingTransac.setBetrag(myBetrag);
		}
		clearingTransac.setZweck(this.kartenNummer);
		// Speichern
		clearingTransac.store();
		saved = true;
	    } catch (final RemoteException e) {
		Logger.error(I18N.tr("ClearingEntry.createSepaTransactionFailed") + e.getMessage(), e); //$NON-NLS-1$
	    } catch (final ApplicationException e) {
		Logger.error(I18N.tr("ClearingEntry.saveSepaTransactionFailed") + e.getMessage(), e); //$NON-NLS-1$
	    }
	}
	return saved;
    }

    /**
     * Setzt das Abbuchungskonto
     *
     * @param konto
     *            das Abbuchungskonto
     * @throws NullArgumentException
     * @throws RemoteException
     */
    public void setKonto(final Konto konto) throws NullArgumentException, RemoteException {

	if (konto == null) {
	    throw new NullArgumentException(I18N.tr("ClearingEntry.clearingAccountIsNull")); //$NON-NLS-1$
	}
	final String ktoId = konto.getMeta(MahoeSynchronizeBackend.CLEARING_ACCOUNT, StringUtils.EMPTY);
	if (StringUtils.isNotEmpty(ktoId)) {
	    this.konto = Utils.getKonto(ktoId);
	}
    }

    /**
     * &Uuml;berpr&uuml;pt die Parameter die zum Erstellen der Klasse angegeben
     * wurden.
     *
     * @throws NullArgumentException
     */
    private void checkParameter() throws NullArgumentException {

	if (this.kartenNummer == null) {
	    throw new NullArgumentException(I18N.tr("ClearingEntry.cardNumberIsNUll")); //$NON-NLS-1$
	}
	if (this.inhaber == null) {
	    throw new NullArgumentException(I18N.tr("ClearingEntry.ownerIsNull")); //$NON-NLS-1$
	}
	if (this.iban == null) {
	    throw new NullArgumentException(I18N.tr("ClearingEntry.ibanIsNull")); //$NON-NLS-1$
	}
	if (this.bic == null) {
	    throw new NullArgumentException(I18N.tr("ClearingEntry.BicIsNull")); //$NON-NLS-1$
	}
	if (this.betrag == null) {
	    throw new NullArgumentException(I18N.tr("ClearingEntry.amountIsNull")); //$NON-NLS-1$
	}
    }

    /**
     * &Uuml;berpr&uuml;ft, ob die Ausgleichsbuchung schon als angelegte
     * Auslandsberweisung (SEPA) existiert.
     *
     * @param hcbiService
     *            der HBCIDBService
     * @return <i>true</i> falls die berweisung schon existiert, <i>false</i>
     *         sonst.
     * @throws RemoteException
     */
    private boolean checkSEPA(final HBCIDBService hcbiService) throws RemoteException {

	Boolean exists = false;
	final DBIterator<AuslandsUeberweisung> transactions = hcbiService.createList(AuslandsUeberweisung.class);
	transactions.setOrder("ORDER BY id DESC"); //$NON-NLS-1$
	// Nur die letzten 20 Tage for dem Flligkeitsdatum berprfen
	final LocalDate faellig = LocalDate.ofEpochDay(this.datum.getTime() / 86400000L).minusDays(30);
	transactions.addFilter("(ausgefuehrt_am is null or ausgefuehrt_am > ?) ", //$NON-NLS-1$
		new Object[] { new java.sql.Date(faellig.toEpochDay() * 86400000L) });
	while (transactions.hasNext()) {
	    final AuslandsUeberweisung sepaBuchung = transactions.next();
	    final String gkInhaber = StringUtils.isNotEmpty(sepaBuchung.getGegenkontoName())
		    ? sepaBuchung.getGegenkontoName() : StringUtils.EMPTY;
	    final double sepaBetrag = sepaBuchung.getBetrag();
	    final String sepaKartenNummer = StringUtils.isNotEmpty(sepaBuchung.getZweck()) ? sepaBuchung.getZweck()
		    : StringUtils.EMPTY;
	    final String myInhaber = Utils.replaceUmlauts(this.inhaber);
	    if (myInhaber.equalsIgnoreCase(gkInhaber) && this.kartenNummer.equals(sepaKartenNummer)
		    && this.betrag.abs().doubleValue() == sepaBetrag) {
		Logger.debug("Equal..."); //$NON-NLS-1$
		exists = true;
		break;
	    } else {
		Logger.debug("Not equal..."); //$NON-NLS-1$
	    }
	}
	return exists;
    }

    /**
     * &Uuml;berpr&uuml;ft, ob die Ausgleichsbuchung schon als ausgefhrte
     * berweisung existiert.
     *
     * @param hcbiService
     *            der HBCIDBService
     * @return <i>true</i>, falls die berweisung schon existiert, <i>false</i>
     *         sonst.
     * @throws RemoteException
     */
    private boolean checkTransactions(final HBCIDBService hcbiService) throws RemoteException {

	Boolean exists = false;
	final DBIterator<Konto> accounts = hcbiService.createList(Konto.class);

	if (this.konto != null) {
	    final String ktoID = this.konto.getID();
	    accounts.addFilter("(id <> " + ktoID + ")"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	while (accounts.hasNext() && !exists) {
	    final Konto account = accounts.next();

	    @SuppressWarnings("unchecked")
	    final DBIterator<Umsatz> transactions = account.getUmsaetze(35);
	    while (transactions.hasNext()) {
		final Umsatz umsatz = transactions.next();
		final String gkInhaber = StringUtils.isNotEmpty(umsatz.getGegenkontoName()) ? umsatz.getGegenkontoName()
			: StringUtils.EMPTY;
		final double umBetrag = umsatz.getBetrag();
		final String umKartenNummer = StringUtils.isNotEmpty(umsatz.getZweck()) ? umsatz.getZweck()
			: StringUtils.EMPTY;
		// Compare
		if (this.inhaber.equalsIgnoreCase(gkInhaber) && this.kartenNummer.equals(umKartenNummer)
			&& this.betrag.abs().doubleValue() == umBetrag) {
		    Logger.debug("Equal..."); //$NON-NLS-1$
		    exists = true;
		    break;
		}
		Logger.debug("Not equal..."); //$NON-NLS-1$
	    }
	}
	return exists;
    }
}
