/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.santander.web;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;

import com.gargoylesoftware.htmlunit.ElementNotFoundException;
import com.gargoylesoftware.htmlunit.FailingHttpStatusCodeException;
import com.gargoylesoftware.htmlunit.WebClient;
import com.gargoylesoftware.htmlunit.html.DomNode;
import com.gargoylesoftware.htmlunit.html.HtmlAnchor;
import com.gargoylesoftware.htmlunit.html.HtmlDivision;
import com.gargoylesoftware.htmlunit.html.HtmlPage;
import com.gargoylesoftware.htmlunit.html.HtmlTable;
import com.gargoylesoftware.htmlunit.html.HtmlTableBody;
import com.gargoylesoftware.htmlunit.html.HtmlTableRow;

import de.frankmuenster.mahoe.MahoeSynchronizeBackend;
import de.frankmuenster.mahoe.banks.santander.SantanderBank;
import de.frankmuenster.mahoe.banks.santander.web.AbstractSantanderWebSite.DocumentType;
import de.frankmuenster.mahoe.statements.AbstractAccountStatementFinder;
import de.frankmuenster.mahoe.statements.MahoeAccountStatement;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.WebUtils;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.rmi.KontoType;
import de.willuhn.logging.Logger;

/**
 * Class to get account statements from the banks web page
 *
 * @author Frank Mnster
 */
@SuppressWarnings("PMD.ExcessiveImports")
public class SantanderAccountStatementFinder extends AbstractAccountStatementFinder<SantanderWebAccountInformation> {

    /**
     * Account number
     */
    private final String accountNumber;

    /**
     * Document type to download
     */
    private final DocumentType docType;

    /**
     * Creates a new SantanderAccountStatement object with a webClient and account information
     *
     * @param webClient
     * @param accountInfo
     * @throws WebSiteException
     */
    public SantanderAccountStatementFinder(final WebClient webClient, final SantanderWebAccountInformation accountInfo)
            throws WebSiteException {

        super(webClient, accountInfo);
        this.accountNumber = this.getAccountNumber();
        this.docType = this.getDocType();

    }

    /**
     * @see AbstractAccountStatementFinder#startSearch()
     */
    @Override
    protected void startSearch() throws WebSiteException {

        HtmlPage page;
        try {

            page = SantanderWebSiteUtils.getDocumentFilteredPage(this.webClient, this.docType);

            if (page.asText().contains("Es liegen zur Zeit keine Nachrichten fr Sie vor")) { //$NON-NLS-1$
                Logger.info(I18N.tr("AccountStatementFinder.noCreditCardInvoice")); //$NON-NLS-1$
            } else {
                // Statements exists --> process the entries and search for
                // additional pages
                this.processStatementPages(page);
            }
        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(I18N.tr("AccountStatementFinder.downloadFailed"), e); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("AccountStatementFinder.downloadFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.WEBSITE_NOT_REACHABLE, e);
        }

    }

    /**
     * Returns the account number to be used
     *
     * @return the account number to be used
     * @throws WebSiteException
     */
    private String getAccountNumber() throws WebSiteException {

        String accNumber = StringUtils.EMPTY;
        try {
            if (KontoType.GIRO.getValue() == this.accountInfo.getKonto().getAccountType()) { // NOPMD
                accNumber = this.accountInfo.getKonto().getKontonummer();
            } else if (KontoType.KREDITKARTE.getValue() == this.accountInfo.getKonto().getAccountType()) {
                accNumber = this.accountInfo.getKonto().getMeta(MahoeSynchronizeBackend.CARD_NUMBER, StringUtils.EMPTY);
            } else {
                throw new WebSiteException(I18N.tr("AccountStatementFinder.accountTypeNotSupported"), //$NON-NLS-1$
                        WebSiteExceptionTypes.ACCOUNT_INFO_WRONG);

            }
        } catch (final RemoteException e) {
            throw new WebSiteException(I18N.tr("AccountStatementFinder.failedGettingAccountNumber"), e); //$NON-NLS-1$
        }
        return accNumber;
    }

    /**
     * Returns the document type to download based on account type
     *
     * @return the {@link DocumentType} to download
     * @throws WebSiteException
     */
    private DocumentType getDocType() throws WebSiteException {

        DocumentType dType = null;
        try {
            if (KontoType.GIRO.getValue() == this.accountInfo.getKonto().getAccountType()) { // NOPMD
                dType = DocumentType.KONTOAUSZUG;
            } else if (KontoType.KREDITKARTE.getValue() == this.accountInfo.getKonto().getAccountType()) {
                dType = DocumentType.KARTENABRECHNUNG;
            } else {
                throw new WebSiteException(I18N.tr("AccountStatementFinder.accountTypeNotSupported"), //$NON-NLS-1$
                        WebSiteExceptionTypes.ACCOUNT_INFO_WRONG);
            }
        } catch (final RemoteException e) {
            throw new WebSiteException(I18N.tr("AccountStatementFinder.failedGettingAccountNumber"), e); //$NON-NLS-1$
        }
        return dType;
    }

    /**
     * Processes row and searches for PDF account statements. If found a {@link MahoeAccountStatement} is created and
     * {@link #addStatement(MahoeAccountStatement, byte[])} is called to add the statement the list of account
     * statements.
     *
     * @param row
     *            the HtmlTableRow to inspect
     * @return <i>true</i> in case all is done, means the defined search period has been inspected, <i>false</i> else
     */
    private boolean processRow(final HtmlTableRow row) {

        final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yy", Locale.GERMAN); //$NON-NLS-1$

        if (row.getCell(2).getTextContent().contains(this.accountNumber)) {

            // entry for account found --> check if PDF sheet and prepare to
            // save
            final String pdfDateString = row.getCell(3).getTextContent().trim().replaceAll("\u00a0", //$NON-NLS-1$
                    StringUtils.EMPTY);
            Logger.debug(I18N.tr("AccountStatementFinder.billingDate", pdfDateString)); //$NON-NLS-1$
            try {
                final Date pdfDate = sdf.parse(pdfDateString);
                if (pdfDate.before(this.lastStatementDate)) {
                    // OK, here we have finished
                    return true; // NOPMD
                }

                final DomNode urlText = row.getCell(1).getFirstChild().getNextSibling();

                Logger.debug("URLtext: " + urlText); //$NON-NLS-1$
                HtmlAnchor anchor = null;
                if (urlText instanceof HtmlAnchor) {
                    anchor = (HtmlAnchor) urlText;
                    Logger.debug("Abrechnung gefunden."); //$NON-NLS-1$
                    final URL fullUrl = row.getHtmlPageOrNull().getFullyQualifiedUrl(anchor.getHrefAttribute());
                    anchor.setAttribute("href", fullUrl.toString()); //$NON-NLS-1$
                    Logger.debug("URL: " + anchor.getHrefAttribute()); //$NON-NLS-1$

                    final byte[] data = WebUtils.getPdfAsByteArray(anchor);
                    final MahoeAccountStatement statementInfo = new MahoeAccountStatement(pdfDate, // NOPMD
                            SantanderBank.NAME, this.docType.getValue());
                    this.addStatement(statementInfo, data);

                } else {
                    Logger.error(I18N.tr("SantanderWebSiteKreditkarte.pdfUrlNotFound")); //$NON-NLS-1$
                }
            } catch (final MalformedURLException e) {
                Logger.error(I18N.tr("SantanderWebSiteKreditkarte.pdfUrlMalforment", //$NON-NLS-1$
                        e.getMessage()), e);
            } catch (final ParseException e) {
                Logger.error(I18N.tr("SantanderWebSiteKreditkarte.parsePdfDateFailed", //$NON-NLS-1$
                        e.getMessage()), e);
            } catch (final RemoteException e) {
                Logger.error(I18N.tr("SantanderWebSiteKreditkarte.savePdfFailed", e.getMessage()), //$NON-NLS-1$
                        e);
            }

        }
        // Row process so not finished yet
        return false;
    }

    /**
     * Process the statements on the actual page and checks for additional pages as well.
     *
     * @param page
     *            the first page to start processing with
     */
    private void processStatementPages(final HtmlPage page) {

        boolean allDone = false;
        HtmlPage actualPage = page;
        do {

            final HtmlDivision content = (HtmlDivision) actualPage.getElementById("content"); //$NON-NLS-1$
            final List<HtmlTable> pdfTables = content.getElementsByAttribute("table", "summary", //$NON-NLS-1$ //$NON-NLS-2$
                    "List of payments"); //$NON-NLS-1$
            if (pdfTables.size() == AbstractSantanderWebSite.MAX_TABLE_SIZE) {
                final HtmlTableBody body = pdfTables.get(0).getBodies().get(0);

                final Iterator<HtmlTableRow> rows = body.getRows().iterator();
                while (rows.hasNext() && !allDone) {
                    allDone = this.processRow(rows.next());
                }
            }
            if (!allDone) {
                // Suche nach weiteren Seiten
                try {
                    final HtmlAnchor nextLink = actualPage.getAnchorByText(AbstractSantanderWebSite.NEXT_LINK_TEXT);
                    actualPage = nextLink.click();
                } catch (final ElementNotFoundException e) {
                    Logger.debug(I18N.tr(AbstractSantanderWebSite.I18N_NEXT_LINK_NOT_FOUND));
                    allDone = true;
                } catch (final IOException e) {
                    Logger.error(I18N.tr("SantanderWebSiteKreditkarte.loadNextPageFailed"), e); //$NON-NLS-1$
                    allDone = true;
                }
            }
        } while (!allDone);

    }
}
