/*******************************************************************************
 * Copyright 2015 - 2019 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.targobank.web; // NOPMD

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;

import com.gargoylesoftware.htmlunit.ElementNotFoundException;
import com.gargoylesoftware.htmlunit.FailingHttpStatusCodeException;
import com.gargoylesoftware.htmlunit.Page;
import com.gargoylesoftware.htmlunit.WebClient;
import com.gargoylesoftware.htmlunit.html.DomElement;
import com.gargoylesoftware.htmlunit.html.DomNode;
import com.gargoylesoftware.htmlunit.html.DomNodeList;
import com.gargoylesoftware.htmlunit.html.HtmlAnchor;
import com.gargoylesoftware.htmlunit.html.HtmlElement;
import com.gargoylesoftware.htmlunit.html.HtmlEmphasis;
import com.gargoylesoftware.htmlunit.html.HtmlInput;
import com.gargoylesoftware.htmlunit.html.HtmlOption;
import com.gargoylesoftware.htmlunit.html.HtmlPage;
import com.gargoylesoftware.htmlunit.html.HtmlSelect;
import com.gargoylesoftware.htmlunit.html.HtmlTable;
import com.gargoylesoftware.htmlunit.html.HtmlTableBody;
import com.gargoylesoftware.htmlunit.html.HtmlTableRow;

import de.frankmuenster.mahoe.banks.targobank.TargoBank;
import de.frankmuenster.mahoe.banks.targobank.pdfextractor.TargoBankPdfExtract;
import de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor;
import de.frankmuenster.mahoe.pdfextractor.InvalidCardNumber;
import de.frankmuenster.mahoe.pdfextractor.PdfBuchung;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.TransactionSaver;
import de.frankmuenster.mahoe.utils.Utils;
import de.frankmuenster.mahoe.web.AbstractWebSite;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.Settings;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/***
 * Statische Implementierungen der Umsatzfunktionen
 *
 * @author Frank Mnster
 */
public class TargoBankWebSiteTransactionsImpl {

    /**
     * Dienst auf Webseite nicht verfgbar
     */
    private static final String CHECK_SERVICE = "Der Dienst ist derzeit nicht verf"; //$NON-NLS-1$

    /**
     * Kennung fr eine Kreditkartenabrechnung
     */
    private static final String EBSCL1 = "EBSCL1"; //$NON-NLS-1$

    /**
     * Minimun cell size of a table row with account sheet PDFs
     */
    private static final int MIN_ROW_CELL_SIZE = 2;

    /**
     * Tabel1 ID Name
     */
    private static final String TABLE1 = "Table1"; //$NON-NLS-1$

    /**
     * Die Kontoinformationen
     */
    private final TargoBankWebAccountInformation accountInfo;

    /**
     * Indikator das letztes PDF erreicht ist.
     */
    private boolean lastPdfToScanHit;

    /**
     * Der WebClient
     */
    private final WebClient webClient;

    /**
     * Liest die vorgemerkten Umstze.
     *
     * @param table
     *            die Table mit dem Umstzen
     * @return eine Liste mit den Umstzen
     */
    private static List<Umsatz> readWebSiteTable(final HtmlTable table) {

        final List<Umsatz> umsaetze = new ArrayList<>();

        // table found enumerate entries
        final List<HtmlTableBody> tbBodies = table.getBodies();
        final SimpleDateFormat sdf = new SimpleDateFormat("dd/MM/yyyy", Locale.getDefault()); //$NON-NLS-1$
        for (final HtmlTableBody tbBody : tbBodies) {
            // step through bodies rows (normally only one)
            final List<HtmlTableRow> tbRows = tbBody.getRows();
            for (final HtmlTableRow tbRow : tbRows) {
                try {
                    final String datumStr = tbRow.getCell(0).getTextContent().trim()
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    Logger.debug(I18N.tr("TargoBankWebSiteTransactionsImpl.transactionDate", datumStr)); //$NON-NLS-1$
                    if ("Datum".equals(datumStr)) { //$NON-NLS-1$
                        continue;
                    }

                    // Neuen Umsatz anlegen
                    final HBCIDBService dbService = Settings.getDBService();
                    final Umsatz umsatz = dbService.createObject(Umsatz.class, null);

                    // Zweck, besteht aus Details, Stadt und Land
                    final String details = tbRow.getCell(1).getTextContent().trim()
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    final String stadt = tbRow.getCell(2).getTextContent().trim()
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    final String land = tbRow.getCell(3).getTextContent().trim()
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);

                    final String zweck = details + " " + stadt + " " //$NON-NLS-1$ //$NON-NLS-2$
                            + land;
                    Utils.loadUmsatz(umsatz, zweck);

                    // Betrag
                    final String betrag = tbRow.getCell(4).getTextContent().trim()
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    umsatz.setBetrag(Utils.convertStringToDouble(betrag));

                    // Fremdwhungsbetrag
                    final String fcValue = tbRow.getCell(5).getTextContent().trim()
                            .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
                    if (StringUtils.isNotEmpty(fcValue)) {
                        umsatz.setKommentar(I18N.tr("TargoBankWebSiteTransactionsImpl.foreignCurrencySaldo", fcValue)); //$NON-NLS-1$
                    }

                    // Buchungsdatum
                    final Date datum = Utils.parseOrNull(sdf, datumStr);
                    umsatz.setDatum(datum);
                    umsatz.setValuta(datum);
                    umsatz.setFlags(Umsatz.FLAG_NOTBOOKED);
                    umsaetze.add(umsatz);
                } catch (final RemoteException re) {
                    Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.createUmsatzError", re.getMessage()), re); //$NON-NLS-1$
                }
            } // end for tbRows
        } // end for tbBodies

        return umsaetze;
    }

    /**
     * Konstruktor
     *
     * @param webClient
     * @param accountInfo
     */
    public TargoBankWebSiteTransactionsImpl(final WebClient webClient,
            final TargoBankWebAccountInformation accountInfo) {

        this.webClient = webClient;
        this.accountInfo = accountInfo;
    }

    /**
     * Liste die gebuchten Umstze ein.
     *
     * @return ein Liste mit den gebuchtenUmstzen
     * @throws WebSiteException
     */
    protected List<Umsatz> getBookedTransactions() throws WebSiteException {

        final List<Umsatz> umsaetze = new ArrayList<>();

        // Zwei Monate weiter zurckgehen, nur um Sicher zu sein
        final Date startDateMod = new Date(
                this.accountInfo.getLastTransactionDate().getTime() - 60 * 24 * 60 * 60 * 1000);

        // ffnen der PDF Liste
        Logger.debug(I18N.tr("TargoBankWebSiteTransactionsImpl.readPDFfor", //$NON-NLS-1$
                Utils.maskCardNummer(this.accountInfo.getCardAccount())));
        HtmlPage pageAbr;
        try {
            pageAbr = this.webClient.getPage(TargoBank.URL_ABBRECHNUNGEN);
            if (pageAbr.asText().contains(TargoBankWebSiteTransactionsImpl.CHECK_SERVICE)) {
                throw new WebSiteException(I18N.tr("TargoBankWebSiteTransactionsImpl.webSiteServiceNotReachable"), //$NON-NLS-1$
                        WebSiteExceptionTypes.SERVICE_UNAVAILABLE);
            } else if (pageAbr.asText().contains("Online-Postfach")) { //$NON-NLS-1$

                // search and open list of account sheets for card account
                final HtmlPage pageNew = getAccountSheetsPage(pageAbr);

                // Search in PDF until given date
                final List<Umsatz> transactions = this.readPDFSites(pageNew, startDateMod);
                // add found transaction to list
                umsaetze.addAll(transactions);

                // Return the list with found transactions
                return umsaetze;

            } else {
                Logger.warn(TargoBankWebSite.PDFSITE_NOT_FOUND);
                throw new WebSiteException(TargoBankWebSite.PDFSITE_NOT_FOUND, WebSiteExceptionTypes.PAGE_NOT_FOUND);
            }

        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.readBookFailedLog") + e.getMessage()); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("TargoBankWebSiteTransactionsImpl.readBookFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.HTTP_STATUS_CODE, e);
        }

    }

    /**
     * Liste die vorgemerkten Umstze ein.
     *
     * @return ein Liste mit den vorgemerkten Umstzen
     * @throws WebSiteException
     * @throws ApplicationException
     */
    protected List<Umsatz> getPreBookTransactions() throws WebSiteException, ApplicationException {

        List<Umsatz> umsaetze = new ArrayList<>();

        try {
            HtmlPage page = TargoBankWebSiteImpl.openAccount(this.webClient, this.accountInfo);

            final DomElement section = page.getElementById("page"); //$NON-NLS-1$
            final DomNodeList<HtmlElement> tables = section.getElementsByTagName(HtmlTable.TAG_NAME);

            do {
                for (final HtmlElement element : tables) {
                    if (element instanceof HtmlTable) {
                        final HtmlTable table = (HtmlTable) element;
                        Logger.debug("Table summery: " + table.getSummaryAttribute()); //$NON-NLS-1$
                        if (table.getSummaryAttribute().equals("Activity Details")) { //$NON-NLS-1$
                            umsaetze = TargoBankWebSiteTransactionsImpl.readWebSiteTable(table);
                        }
                    }
                }
                page = this.existsNextPage();
            } while (page != null);

        } catch (final WebSiteException e) {
            Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.readPreBookFailed", e.getMessage())); //$NON-NLS-1$
        }
        // Salden hinzufgen
        double zwischenSaldo = this.accountInfo.getSaldo().doubleValue();
        for (final Umsatz umsatz : umsaetze) {
            try {
                umsatz.setSaldo(zwischenSaldo);
                zwischenSaldo = Utils.roundNumber(zwischenSaldo - umsatz.getBetrag(), 2);
            } catch (final RemoteException e) {
                Logger.warn(I18N.tr("TargoBankWebSiteTransactionsImpl.calculatingSaldoFailed", e.getMessage())); //$NON-NLS-1$
            }
        }

        return umsaetze;

    }

    /**
     * Check if a next button exits and load the page
     *
     * @return next page or <code>null</code>
     */
    private HtmlPage existsNextPage() {

        // check for next page
        // TODO: Test auf weitere Tabellenseiten
        // Konnte noch nicht erfolgen, da ich zu wenig
        // Umsaetze habe.

        return null;
    }

    /**
     * Liest alle Umstze aus den PDF der Webseite bis zum angegebenem datum.
     *
     * @param page
     *            die HtmlUnit Webseite
     * @param startDate
     *            das Startdatum
     * @return List emit den gefundenen Umstzen
     */
    private List<Umsatz> findPDFsInTable(final HtmlPage page, final Date startDate) {

        final List<Umsatz> umsaetze = new ArrayList<>();
        final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy", Locale.GERMAN); //$NON-NLS-1$

        final HtmlTable table = (HtmlTable) page.getElementById(TargoBankWebSiteTransactionsImpl.TABLE1);
        final DomNodeList<HtmlElement> tables = table.getElementsByTagName(HtmlTable.TAG_NAME);

        final HtmlTableBody tbBody = ((HtmlTable) tables.item(1)).getBodies().get(0);

        // step through bodies rows (normally only one)
        final List<HtmlTableRow> tbRows = tbBody.getRows();
        final Iterator<HtmlTableRow> tbRowsIt = tbRows.iterator();
        HtmlTableRow tbRow;
        while (tbRowsIt.hasNext() && !this.lastPdfToScanHit) {
            // take next row
            tbRow = tbRowsIt.next();

            // check if all cells are present
            if (tbRow.getCells().size() < TargoBankWebSiteTransactionsImpl.MIN_ROW_CELL_SIZE) {
                continue;
            }
            // get date
            final String pdfDateString = tbRow.getCell(1).getTextContent().trim()
                    .replaceAll(AbstractWebSite.NON_BREAK_SPACE, StringUtils.EMPTY);
            Logger.debug(I18N.tr("TargoBankWebSiteTransactionsImpl.billingDate", pdfDateString)); //$NON-NLS-1$

            try {
                final Date pdfDate = sdf.parse(pdfDateString);
                if (pdfDate.before(startDate)) {
                    Logger.debug(I18N.tr("TargoBankWebSiteTransactionsImpl.reachedLastPdf")); //$NON-NLS-1$
                    this.lastPdfToScanHit = true;
                } else {
                    final DomNode urlText = tbRow.getCell(2).getFirstChild();
                    Logger.debug("URLtext: " + urlText); //$NON-NLS-1$

                    final HtmlAnchor anchor = this.getAnchorFromUrl(urlText, page);
                    if (anchor != null) {
                        final List<Umsatz> transactions = this.loadTransactionsFromPDF(anchor);
                        umsaetze.addAll(transactions);
                    }
                }
            } catch (final ParseException e1) {
                Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.readPdfDateFailed", e1.getMessage()), e1); //$NON-NLS-1$
            }

        } // end for tbRows

        return umsaetze;
    }

    /**
     * Search the open document page for the current account and returns the
     * page with the list of account sheets only from this card account.
     *
     * @param pageAbr
     *            page with open the account sheets main document page
     * @return page with dedicated account sheets for the actual card account
     * @throws WebSiteException
     *             in case the account could not be found
     * @throws IOException
     */
    private HtmlPage getAccountSheetsPage(final HtmlPage pageAbr) throws WebSiteException, IOException {

        Logger.debug(pageAbr.asText());

        // get card account number and remove spaces
        final String cardAccount = this.accountInfo.getCardAccount().replaceAll(" ", StringUtils.EMPTY); //$NON-NLS-1$

        try {
            final HtmlSelect selIndex1 = pageAbr.getElementByName("SelIndex1"); //$NON-NLS-1$

            Logger.debug(selIndex1.asXml());

            // Search card account in pull down list
            final List<HtmlOption> options = selIndex1.getOptions();
            HtmlOption option = null;
            for (final HtmlOption opt : options) {
                if (opt.getValueAttribute().contains(cardAccount)) {
                    option = opt;
                    break;
                }
            }

            // Card account found? If not throw exception
            if (option == null) {
                Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.cardAccountNotFound")); //$NON-NLS-1$ //NOPMD
                throw new WebSiteException(I18N.tr("TargoBankWebSiteTransactionsImpl.cardAccountNotFound"), //$NON-NLS-1$
                        WebSiteExceptionTypes.KONTO_NOT_FOUND);
            }
            // select card account
            selIndex1.setSelectedAttribute(option, true);

            // get and press submit button
            final HtmlInput submit = (HtmlInput) pageAbr.getElementById("btnRechercher"); //$NON-NLS-1$

            // return the page
            return submit.click();
        } catch (final ElementNotFoundException e) {
            Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.AccountsDropdownNotFound"), e); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("TargoBankWebSiteTransactionsImpl.AccountsDropdownNotFound"), //$NON-NLS-1$
                    WebSiteExceptionTypes.PAGE_NOT_FOUND, e);

        }
    }

    /**
     * Erstellt den Anker aus einer URL DOM Node abhngig davon, ob das Dokument
     * schon gelesen wurde oder nicht. Nicht Kreditkartenabrechnungen werden
     * heraus gefiltert.
     *
     * @param urlText
     *            der <b>DomNode</b> mit dem Dokumentenlink
     * @param page
     * @return der <b>HMTLAnchor</b> fr den Download des Dokuments
     */
    private HtmlAnchor getAnchorFromUrl(final DomNode urlText, final HtmlPage page) {

        HtmlAnchor anchor = null;
        if (urlText.getTextContent().toUpperCase(Locale.getDefault())
                .contains(TargoBankWebSiteTransactionsImpl.EBSCL1)) {

            if (urlText instanceof HtmlAnchor) {
                anchor = (HtmlAnchor) urlText;
                Logger.debug("Alte Abrechnung gefunden."); //$NON-NLS-1$
            } else if (urlText instanceof HtmlEmphasis) {
                Logger.debug("Neue Abrechnung gefunden."); //$NON-NLS-1$
                anchor = (HtmlAnchor) urlText.getFirstChild();
            } else {
                Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.pdfUrlNotFound")); //$NON-NLS-1$
            }
        } else {
            Logger.debug("Ist keine Kreditkartenabrechnung"); //$NON-NLS-1$
        }

        // compose the full URL in case an anchor has been found
        if (anchor != null) {
            URL fullUrl;
            try {
                fullUrl = page.getFullyQualifiedUrl(anchor.getHrefAttribute());
                anchor.setAttribute("href", fullUrl.toString()); //$NON-NLS-1$
                Logger.debug("URL: " + anchor.getHrefAttribute()); //$NON-NLS-1$
            } catch (final MalformedURLException e) {
                anchor = null; // NOPMD
                Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.pdfUrlMalformend", //$NON-NLS-1$
                        e.getMessage()), e);
            }
        }

        return anchor;
    }

    /**
     * Lde alle Umstze aus dem im Anchor angegebenen PDF
     *
     * @param anchor
     *            zu dem PDF.
     * @return Liste mit allen gefundenen Umstzen. Kann leer sein.
     */
    private List<Umsatz> loadTransactionsFromPDF(final HtmlAnchor anchor) {

        List<Umsatz> transactions = new ArrayList<>();

        try {

            final Page pdf = this.webClient.getPage(anchor.getHrefAttribute());

            if (pdf.isHtmlPage()) {
                Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.loadPdfFailed", //$NON-NLS-1$
                        anchor.getHrefAttribute()));
            } else {

                Logger.debug("Lade PDF"); //$NON-NLS-1$
                final InputStream stream = pdf.getWebResponse().getContentAsStream();
                final String cardAccount = this.accountInfo.getCardAccount().replaceAll(" ", StringUtils.EMPTY); //$NON-NLS-1$
                Logger.debug(I18N.tr("TargoBankWebSiteTransactionsImpl.extractTransactions", cardAccount)); //$NON-NLS-1$
                final AbstractPdfExtractor extractor = new TargoBankPdfExtract(stream, cardAccount);
                final List<PdfBuchung> buchungenPDF = extractor.getExtract();

                // Ressourcen schlieen
                extractor.close();
                stream.close();

                transactions = TransactionSaver.getTransactions(buchungenPDF);

            }
        } catch (IOException | InvalidCardNumber e) {
            Logger.error(I18N.tr("TargoBankWebSiteTransactionsImpl.readPdfContentFailed", e.getMessage()), e); //$NON-NLS-1$
        }

        return transactions;
    }

    /**
     * Ld alle Seiten mit PDF Links und wertet diese aus, soweit wie ntig.
     *
     * @param page
     * @param startDate
     * @return die Liste mit allen gefundenen Umstzen
     * @throws IOException
     */
    private List<Umsatz> readPDFSites(final HtmlPage page, final Date startDate) throws IOException {

        final List<Umsatz> umsaetze = new ArrayList<>();

        boolean hasNext;
        this.lastPdfToScanHit = false;

        HtmlPage pageInternal = page;

        do {

            final List<Umsatz> transactions = this.findPDFsInTable(pageInternal, startDate);
            umsaetze.addAll(transactions);

            hasNext = false;
            if (!this.lastPdfToScanHit) {
                final HtmlTable table = (HtmlTable) pageInternal
                        .getElementById(TargoBankWebSiteTransactionsImpl.TABLE1);
                final DomNodeList<HtmlElement> inputButtons = table.getElementsByTagName(HtmlInput.TAG_NAME);
                for (final HtmlElement htmlElement : inputButtons) {
                    if (htmlElement.getId().equals("NEXT")) { //$NON-NLS-1$
                        hasNext = true;
                        pageInternal = htmlElement.click();
                    }
                }
            }

        } while (!this.lastPdfToScanHit && hasNext);

        return umsaetze;

    }
}
