/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.jobs;

import java.rmi.RemoteException;
import java.util.List;

import de.frankmuenster.mahoe.messages.AbstractBankMessages;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.web.AbstractWebSite;
import de.frankmuenster.mahoe.web.WebAccountInformation;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.willuhn.jameica.hbci.SynchronizeOptions;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.jameica.hbci.synchronize.jobs.SynchronizeJob;
import de.willuhn.jameica.hbci.synchronize.jobs.SynchronizeJobKontoauszug;
import de.willuhn.logging.Level;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;
import de.willuhn.util.ProgressMonitor;

/**
 * Abstrakte Klasse fr Kontoauszugsabfrage fr Mahoe Konten. Hier verarbeiten
 * wir hauptschlich die Fehlerbehandlung.
 *
 * @author Frank Mnster
 * @param <I>
 *            die bankspezifischen WebAccountInformation
 * @param <M>
 *            die bankspezifische AbstractBankMessages
 */
public abstract class AbstractMahoeSynchronizeJobKontoauszug<I extends WebAccountInformation,
        M extends AbstractBankMessages> extends SynchronizeJobKontoauszug implements InterfaceMahoeSynchronizeJob {

    /**
     * Das Konto
     */
    protected transient Konto konto;

    /**
     * Die Kundennummer. Extra damit man nicht jedes mal die RemoteException
     * abfangen muss.
     */
    protected transient String kundennummer;

    /**
     * Das Webseiten Objekt
     */
    protected transient AbstractWebSite<I, M> webSite;

    /**
     * Kontoinformationen
     */
    protected transient I accountInfo;

    /**
     * Das Progessmonitor Objekt
     */
    protected transient ProgressMonitor monitor;

    /**
     * Standardkonstruktor
     */
    public AbstractMahoeSynchronizeJobKontoauszug() {

        super();
    }

    /**
     * @see InterfaceMahoeSynchronizeJob#execute(ProgressMonitor, boolean)
     */
    @Override
    public void execute(final ProgressMonitor monitor, final boolean lastJob) throws Exception { // NOPMD

        this.konto = (Konto) this.getContext(SynchronizeJob.CTX_ENTITY);
        this.monitor = monitor;

        String accountname;
        try {
            this.kundennummer = this.konto.getKundennummer();
            accountname = this.konto.getBezeichnung();
        } catch (final RemoteException e) {
            throw new ApplicationException(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readAccountInfoFailed"), e); //$NON-NLS-1$
        }
        Logger.info(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readTransactionLoginfo", accountname, //$NON-NLS-1$
                this.kundennummer));

        // Wir laden das BankSeiten-Objekt
        this.getWebSite();
        monitor.setPercentComplete(10);

        // Nur Anmelden, wenn die Verbindung nicht schon besteht
        if (this.webSite.isStillLoggedIn()) {
            monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.connectionStillValid")); //$NON-NLS-1$
        } else {
            monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.startLogin")); //$NON-NLS-1$
            this.login();
            Logger.info(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.loginSuccessful")); //$NON-NLS-1$
            monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.loginSuccessful")); //$NON-NLS-1$
        }
        monitor.setPercentComplete(20);

        // Kontoinformationen
        monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.updateAccountInfo")); //$NON-NLS-1$
        this.accountInfo = this.webSite.getAccountInformation();
        this.accountInfo.saveInformationInAccount(monitor);
        monitor.setPercentComplete(30);

        // Umstze
        monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readTransactions")); //$NON-NLS-1$
        this.handleTransactions();
        monitor.setPercentComplete(60);

        // Banknachrichten
        try {
            final SynchronizeOptions options = new SynchronizeOptions(this.konto);
            if (options.getSyncMessages()) {

                monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readMessages")); //$NON-NLS-1$
                final int savedMsg = this.webSite.getMessages(this.accountInfo);
                if (savedMsg == 0) {
                    monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.noNewMessages")); //$NON-NLS-1$
                } else {
                    monitor.setStatusText(
                            I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.newMessagesStored", savedMsg)); //$NON-NLS-1$
                }
            }
        } catch (WebSiteException | ApplicationException | RemoteException e) {
            monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readMessagesFailed")); //$NON-NLS-1$
            Logger.error(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readMessagesFailedLog", e.getMessage()), e); //$NON-NLS-1$
        }

        // Falls letzter Job fr die Bank, abmelden
        if (lastJob) {
            monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.reachedLastJob")); //$NON-NLS-1$
            this.webSite.logoff();
        }
        monitor.setPercentComplete(100);

    }

    /**
     * Log transaction in the given list, but only in case log level is
     * <code>DEBUG</code>
     *
     * @param transactions
     */
    protected void debugLogTransactions(final List<Umsatz> transactions) {

        // Debug-Ausgabe
        if (Logger.isLogging(Level.DEBUG)) {
            for (final Umsatz umsatz : transactions) {
                String umsatzString;
                try {
                    umsatzString = I18N.tr("SynchronizeJobKontoauszug.transactionInfoLog", //$NON-NLS-1$
                            umsatz.getDatum().toString(), umsatz.getZweck(), umsatz.getZweck2(),
                            Double.toString(umsatz.getBetrag()), Double.toString(umsatz.getSaldo()),
                            umsatz.getKommentar());
                    Logger.debug(umsatzString);
                } catch (final RemoteException e) {
                    Logger.error(I18N.tr("SynchronizeJobKontoauszug.createTransactionInfoStringFailed"), e); //$NON-NLS-1$
                }
            }
        }
    }

    /**
     * Liefert das Webseiten-Objekt fr die Bank.
     *
     * @throws WebSiteException
     * @throws ApplicationException
     */
    protected abstract void getWebSite() throws WebSiteException, ApplicationException;

    /**
     * Liest und speichert die Umstze. Ob auch vorgemerkte Umstze bearbeitet
     * werden, entscheidet die konkrete Implementierung.
     *
     * @throws WebSiteException
     * @throws ApplicationException
     */
    protected abstract void handleTransactions() throws WebSiteException, ApplicationException;

    /**
     * Meldet das Webseiten-Objekt an der Bank an.
     *
     * @throws WebSiteException
     * @throws ApplicationException
     */
    protected abstract void login() throws WebSiteException, ApplicationException;

}
