/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.utils;

import java.io.ByteArrayInputStream;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.StandardCharsets;
import java.security.GeneralSecurityException;
import java.util.Arrays;

import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.Base64;

/**
 * Mahoe Crypto-Utility. Uses Jameica en-/decryption methods
 *
 * @author Frank Mnster
 * @version 2.0
 */
public final class Crypto {

    /**
     * Entschlsselt einen verschlsselten Text.
     *
     * @param encryptedText
     *            der verschlsselte Text
     * @return char[] der entschlsselte Text.
     * @throws GeneralSecurityException
     *             , falls irgendwas schief geht.
     */
    public static char[] decrypt(final String encryptedText) throws GeneralSecurityException {

        if (encryptedText == null) {
            throw new GeneralSecurityException(I18N.tr("Crypto.argumentCannotBeNull")); //$NON-NLS-1$
        }

        Logger.debug("decrypting database password"); //$NON-NLS-1$
        try {
            // create bis and bos streams
            final ByteArrayInputStream bis = new ByteArrayInputStream(Base64.decode(encryptedText));
            final ClearableByteArrayOutputStream bos = new ClearableByteArrayOutputStream();
            // decrypt the string
            Application.getSSLFactory().decrypt(bis, bos);
            final char[] decrypted = Crypto.bytesToChars(bos.toByteArray());
            // clear buffer
            bos.clear();
            return decrypted;
        } catch (final Exception e) { // NOPMD
            throw new GeneralSecurityException(I18N.tr("Crypto.decrpytError"), e); //$NON-NLS-1$
        }

    }

    /**
     * Verschlsselt einen Text als char array im Default encoding.
     *
     * @param plaintext
     *            char[] der zu verschlsselnde Text
     * @return String der verschlsselte Text
     * @throws GeneralSecurityException
     *             falls irgend ein Fehler auftritt
     */
    public static String encrypt(final char[] plaintext) // NOPMD
            throws GeneralSecurityException {

        if (plaintext == null) {
            throw new GeneralSecurityException(I18N.tr("Crypto.argumentCannotBeNull")); //$NON-NLS-1$
        }

        Logger.debug("encrypting password with system certificate"); //$NON-NLS-1$
        // Get byte array from char array
        final byte[] byteArray = Crypto.charsToBytes(plaintext.clone());
        Logger.debug("Bytes: " + byteArray); //$NON-NLS-1$
        final ByteArrayInputStream bis = new ByteArrayInputStream(byteArray);
        final ClearableByteArrayOutputStream bos = new ClearableByteArrayOutputStream();
        try {
            Application.getSSLFactory().encrypt(bis, bos);
            Arrays.fill(byteArray, (byte) 0);
            final String encypted = Base64.encode(bos.toByteArray());
            bos.clear();
            return encypted;
        } catch (final Exception e) { // NOPMD
            throw new GeneralSecurityException(I18N.tr("Crypto.encrpytError"), e); //$NON-NLS-1$
        }

    }

    /**
     * Convert a byte array to a character array using UTF-8 with a intermediate
     * string.
     *
     * @param bytes
     *            the byte array
     * @return the character array
     */
    private static char[] bytesToChars(final byte[] bytes) {

        // warp a CharBuffer
        final ByteBuffer byteBuffer = ByteBuffer.wrap(bytes);
        // encode to byte buffer
        final CharBuffer charBuffer = StandardCharsets.UTF_8.decode(byteBuffer);
        // wipe the original char array
        Arrays.fill(bytes, (byte) 0);
        // clear byteBuffer as it is not needed any more
        Crypto.clearByteBuffer(byteBuffer);
        // create the final byte array
        final char[] result = Arrays.copyOf(charBuffer.array(), charBuffer.limit());
        // clear charBuffer as it is not needed any more
        Crypto.clearCharBuffer(charBuffer);
        // return the byte array
        return result;
    }

    /**
     * Convert a character array to a byte array using UTF-8 without a
     * intermediate string. It also wipes the give char array
     *
     * @param chars
     *            the character array
     * @return the byte array
     */
    private static byte[] charsToBytes(final char[] chars) { // NOPMD

        // warp a CharBuffer
        final CharBuffer charBuffer = CharBuffer.wrap(chars);
        // encode to byte buffer
        final ByteBuffer byteBuffer = StandardCharsets.UTF_8.encode(charBuffer);
        // wipe the original char array
        Arrays.fill(chars, (char) 0);
        // clear charBuffer as it is not needed any more
        Crypto.clearCharBuffer(charBuffer);
        // create the final byte array
        final byte[] result = Arrays.copyOf(byteBuffer.array(), byteBuffer.limit());
        // clear byteBuffer as it is not needed any more
        Crypto.clearByteBuffer(byteBuffer);
        // return the byte array
        return result;
    }

    /**
     * Wipes a CharBuffer with zero.
     *
     * @param byteBuffer
     */
    private static void clearByteBuffer(final ByteBuffer byteBuffer) {

        byteBuffer.clear();
        while (byteBuffer.hasRemaining()) {
            byteBuffer.put((byte) 0);
        }

    }

    /**
     * Wipes a CharBuffer with zero.
     *
     * @param charBuffer
     */
    private static void clearCharBuffer(final CharBuffer charBuffer) {

        charBuffer.clear();
        while (charBuffer.hasRemaining()) {
            charBuffer.put((char) 0);
        }

    }

    private Crypto() {

        throw new AssertionError("Instantiating utility class..."); //$NON-NLS-1$
    }

}
