/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.web;

import java.math.BigDecimal;
import java.rmi.RemoteException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;

import de.frankmuenster.mahoe.utils.I18N;
import de.willuhn.jameica.hbci.messaging.ObjectChangedMessage;
import de.willuhn.jameica.hbci.messaging.SaldoMessage;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;
import de.willuhn.util.ProgressMonitor;

/**
 * Minimale Kontoinformationen, die jedes Job auslesen muss.
 *
 * @author Frank Mnster
 */
public abstract class AbstractWebAccountInformation implements WebAccountInformation {

    /**
     * BIC
     */
    private String bic;

    /**
     * Der Kommentar
     */
    private String comment;

    /**
     * IBAN
     */
    private String iban;

    /**
     * Das letzte Umsatzdatum
     */
    private Date lastTransactionDate; // NOPMD

    /**
     * Das Saldo, wie auf der Webseite angezeigt
     */
    private BigDecimal saldo;

    /**
     * Verfgbarer Betrag
     */
    private BigDecimal saldoAvailable;

    /**
     * Das zugrundeliegende Konto
     */
    protected Konto konto;

    /**
     * Der Kontoauszugs-Speicherpfad
     */
    private String savePath;

    /**
     * Standardkonstruktor
     */
    public AbstractWebAccountInformation() {

        // Absichtlich leer
    }

    /**
     * @return the bic
     */
    public String getBic() {

        return this.bic;
    }

    /**
     * @return the comment
     */
    @Override
    public String getComment() {

        return this.comment;
    }

    /**
     * @return the iban
     */
    public String getIban() {

        return this.iban;
    }

    /**
     * @return the konto
     */
    @Override
    public Konto getKonto() {

        return this.konto;
    }

    /**
     * @return the lastTransactionDate
     */
    @Override
    public Date getLastTransactionDate() {

        return this.lastTransactionDate;
    }

    /**
     * Letztes Umsatzdatum schn formatiert
     */
    @Override
    public String getLastTransactionDateAsString() {

        final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy", //$NON-NLS-1$
                Locale.GERMAN);
        return sdf.format(this.lastTransactionDate);
    }

    /**
     * @return the saldo
     */
    @Override
    public BigDecimal getSaldo() {

        return this.saldo;
    }

    /**
     * @return the saldoAvailable
     */
    @Override
    public BigDecimal getSaldoAvailable() {

        return this.saldoAvailable;
    }

    /**
     * @return the savePath
     */
    @Override
    public String getSavePath() {

        return this.savePath;
    }

    /**
     * @see de.frankmuenster.mahoe.web.WebAccountInformation#saveInformationInAccount(de.willuhn.util.ProgressMonitor)
     */
    @Override
    public void saveInformationInAccount(final ProgressMonitor monitor) // NOPMD
            throws ApplicationException {

        if (this.konto == null) {
            throw new ApplicationException(I18N.tr("AbstractWebAccountInformation.kontoIsNull")); //$NON-NLS-1$
        }

        // Saldo
        if (this.saldo != null) {
            monitor.setStatusText(I18N.tr("AbstractWebAccountInformation.updateSaldo")); //$NON-NLS-1$
            try {
                this.konto.setSaldo(this.saldo.doubleValue());
            } catch (final RemoteException e) {
                Logger.error(I18N.tr("AbstractWebAccountInformation.setSaldoFailed"), e); //$NON-NLS-1$
                throw new ApplicationException(I18N.tr("AbstractWebAccountInformation.setSaldoFailed"), e); //$NON-NLS-1$
            }
        }

        // Verfgbarer Betrag
        if (this.saldoAvailable != null) {
            monitor.setStatusText(I18N.tr("AbstractWebAccountInformation.updateSaldoAvailable")); //$NON-NLS-1$
            try {
                this.konto.setSaldoAvailable(this.saldoAvailable.doubleValue());
            } catch (final RemoteException e) {
                Logger.error(I18N.tr("AbstractWebAccountInformation.setSaldoAvailableFailed"), e); //$NON-NLS-1$
                throw new ApplicationException(I18N.tr("AbstractWebAccountInformation.setSaldoAvailableFailed"), e); //$NON-NLS-1$
            }
        }

        // Kommentar
        if (this.comment != null) {
            monitor.setStatusText(I18N.tr("AbstractWebAccountInformation.updateComment")); //$NON-NLS-1$
            try {
                this.konto.setKommentar(this.comment);

            } catch (final RemoteException e) {
                Logger.error(I18N.tr("AbstractWebAccountInformation.updateCommentFailed"), e); //$NON-NLS-1$
                throw new ApplicationException(I18N.tr("AbstractWebAccountInformation.updateCommentFailed"), e); //$NON-NLS-1$
            }
        }

        // IBAN
        if (this.iban != null) {
            monitor.setStatusText(I18N.tr("AbstractWebAccountInformation.updateIban")); //$NON-NLS-1$
            try {
                this.konto.setIban(this.iban.replaceAll("\\s", StringUtils.EMPTY)); //$NON-NLS-1$
            } catch (final RemoteException e) {
                Logger.error(I18N.tr("AbstractWebAccountInformation.updateIbanFailed"), e); //$NON-NLS-1$
                throw new ApplicationException(I18N.tr("AbstractWebAccountInformation.updateIbanFailed"), e); //$NON-NLS-1$
            }
        }

        // BIC
        if (this.bic != null) {
            monitor.setStatusText(I18N.tr("AbstractWebAccountInformation.updateBic")); //$NON-NLS-1$
            try {
                this.konto.setBic(this.bic.replaceAll("//s", StringUtils.EMPTY)); //$NON-NLS-1$
            } catch (final RemoteException e) {
                Logger.error(I18N.tr("AbstractWebAccountInformation.updateBicFailed"), e); //$NON-NLS-1$
                throw new ApplicationException(I18N.tr("AbstractWebAccountInformation.updateBicFailed"), e); //$NON-NLS-1$
            }
        }

        // Informationen von des Sub-Klassen speichern
        try {
            this.saveBankSpecificInfo(monitor);
        } catch (final ApplicationException e) {
            Logger.error(I18N.tr("AbstractWebAccountInformation.setSpezificAccountInfoFailed"), e); //$NON-NLS-1$
            throw new ApplicationException(I18N.tr("AbstractWebAccountInformation.setSpezificAccountInfoFailed"), //$NON-NLS-1$
                    e);
        }

        try {
            this.konto.store();
            Application.getMessagingFactory().sendMessage(new SaldoMessage(this.konto));
            Application.getMessagingFactory().sendMessage(new ObjectChangedMessage(this.konto));
        } catch (final RemoteException e) {
            Logger.error(I18N.tr("AbstractWebAccountInformation.storeKontoFailed"), e); //$NON-NLS-1$
            throw new ApplicationException(I18N.tr("AbstractWebAccountInformation.storeKontoFailed"), e); //$NON-NLS-1$
        }

    }

    /**
     * @param bic
     *            the bic to set
     */
    public void setBic(final String bic) {

        this.bic = bic;
    }

    /**
     * @param comment
     *            the comment to set
     */
    @Override
    public void setComment(final String comment) {

        this.comment = comment;
    }

    /**
     * @param iban
     *            the iban to set
     */
    public void setIban(final String iban) {

        this.iban = iban;
    }

    /**
     * @param konto
     *            the konto to set
     */
    @Override
    public void setKonto(final Konto konto) {

        this.konto = konto;
    }

    /**
     * @param lastTransactionDate
     *            the lastTransactionDate to set
     */
    @Override
    public void setLastTransactionDate(final Date lastTransactionDate) { // NOPMD

        this.lastTransactionDate = lastTransactionDate;
    }

    /**
     * @param saldo
     *            the saldo to set
     */
    @Override
    public void setSaldo(final BigDecimal saldo) {

        this.saldo = saldo;
    }

    /**
     * @param saldoAvailable
     *            the saldoAvailable to set
     */
    @Override
    public void setSaldoAvailable(final BigDecimal saldoAvailable) {

        this.saldoAvailable = saldoAvailable;
    }

    /**
     * @param savePath
     *            the savePath to set
     */
    @Override
    public void setSavePath(final String savePath) {

        this.savePath = savePath;
    }

    /**
     * Hier werden die webseitenspezifischen Kontoinformationen verarbeitet. Muss von der Child-Klasse implementiert
     * werden.<br/>
     * Ein speichern des Kontos ist nicht erforderlich. das macht die Super-Klasse.
     *
     * @param monitor
     * @throws ApplicationException
     */
    protected abstract void saveBankSpecificInfo(ProgressMonitor monitor) throws ApplicationException;

}
